import { NOOP } from 'react-event-base/ObjectUtils';
import { EventModel } from 'react-event-base/Core';
import { createEventTypes } from 'react-event-base/EventUtils';

class BaseCommand {
  constructor() {
    this.requiredArgs = [];

    this.onPreExecute = () => {
      this.firstPreExecute();
      this.onPreExecute = NOOP;
    };

    this.onPostExecute = () => {
      this.firstPostExecute();
      this.onPostExecute = NOOP;
    };
  }

  init(args) {
    this.onBeforeInit();
    this.requiredArgs.forEach(argKey => {
      if (!(argKey in args)) {
        console.warn("[Command:".concat(this.name, "] Argument missing: ").concat(argKey, " while initializing command."));
      }
    });
    Object.assign(this, args);
    this.onAfterInit();
  }

  onBeforeInit() {}

  onAfterInit() {}

  initializeRequiredArgs() {
    return [];
  }

  execute() {
    this.onPreExecute();
    var result = this.doExecute();
    this.onPostExecute();
    return result;
  }

  firstPreExecute() {}

  firstPostExecute() {}

  executeAfterRedo() {
    this.doExecuteAfterRedo();
  }

  unExecute() {
    this.doUnexecute();
  }

  unExecuteBeforeUndo() {
    this.doUnExecuteBeforeUndo();
  }

  doExecute() {}

  doExecuteAfterRedo() {}

  doUnexecute() {}

  doUnExecuteBeforeUndo() {}

  dispose() {}

  getMessage() {}

}

class MacroCommand extends BaseCommand {
  constructor(commands) {
    super();
    this.commands = commands || [];
    this.isMacro = true;
  }

  doExecute() {
    for (var i = 0; i < this.commands.length; i++) {
      this.commands[i].execute();
    }
  }

  doUnexecute() {
    for (var i = this.commands.length - 1; i >= 0; i--) {
      this.commands[i].unExecute();
    }
  }

}

var Events = createEventTypes(['Undo', 'Redo', 'Execute', 'Clear', 'Reload']);

class CommandManager extends EventModel {
  constructor() {
    super(...arguments);
    this.undoArray = [];
    this.redoArrray = [];
  }

  getMaxStackSize() {
    return this.maxStackSize || 999;
  }

  getId() {
    return this.id;
  }

  pushCommand(command) {
    var maxStackSize = this.getMaxStackSize();
    this.undoArray.push(command);
    this.redoArrray.splice(0, this.redoArrray.length);

    if (this.undoArray.length > maxStackSize) {
      this.undoArray.splice(0, this.undoArray.length - maxStackSize);
    }

    this.trigger(Events.Execute, [command]);
  }

  execute(command) {
    var result = command.execute();
    this.pushCommand(command);
    return result;
  }

  canUndo() {
    return this.undoArray.length > 0;
  }

  canRedo() {
    return this.redoArrray.length > 0;
  }

  undo() {
    var lastCmd = this.redoArrray[this.redoArrray.length - 1];

    if (lastCmd) {
      lastCmd.unExecuteBeforeUndo();
    }

    try {
      var command = this.undoArray.pop();

      if (command) {
        this.redoArrray.push(command);
        command.unExecute();
        this.trigger(Events.Undo, [command]);
      }
    } catch (err) {
      console.error(err);
    }
  }

  getUndoSteps() {
    return this.undoArray.length;
  }

  redo() {
    var lastCmd = this.redoArrray[this.redoArrray.length - 2];

    if (lastCmd) {
      lastCmd.executeAfterRedo();
    }

    try {
      var command = this.redoArrray.pop();

      if (command) {
        this.undoArray.push(command);
        command.execute();
        this.trigger(Events.Redo, [command]);
      }
    } catch (err) {
      console.error(err);
    }
  }

  getRedoSteps() {
    return this.redoArrray.length;
  }

  clear() {
    this.undoArray.splice(0, this.undoArray.length);
    this.redoArrray.splice(0, this.redoArrray.length);
    this.trigger(Events.Clear);
  }

  destroy() {
    this.undoArray = undefined;
    this.redoArrray = undefined;
  }

  dump() {}

  reload(commandManager) {
    this.undoArray.splice.apply(this.undoArray, [0, this.undoArray.length].concat(commandManager.undoArray));
    this.redoArrray.splice.apply(this.redoArrray, [0, this.redoArrray.length].concat(commandManager.redoArrray));
    this.trigger(Events.Reload);
  }

}

class AbstractCommandFactory extends EventModel {
  constructor(data) {
    super(data);
    this.cmdTypes = this.cmdTypes || {};
    this.extraCmdArgs = this.extraCmdArgs || {};
  }

  createCommand(cmdType, args) {
    var Cmd = this.cmdTypes[cmdType];

    if (Cmd) {
      var cmd = new Cmd();
      cmd.init(Object.assign(args || {}, this.extraCmdArgs));
      return cmd;
    }
  }

  initializeCommands(cmdTypes) {
    this.cmdTypes = cmdTypes || {};
  }

}
/*
class CommandFactory extends AbstractCommandFactory {
	
	constructor(data) {
		super(data);
		this.initializeCommands({
			
		});
	}

}
*/

export { AbstractCommandFactory, BaseCommand, CommandManager, Events as CommandManagerEvents, MacroCommand };
